@@ -1,10 +1,10 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
 """
 Term based tool to view *colored*, *incremental* diff in a *Git/Mercurial/Svn*
 workspace or from stdin, with *side by side* and *auto pager* support. Requires
-python (>= 2.5.0) and ``less``.
+python (>= 2.6.0) and ``less``.
 """
 
 import sys
@@ -12,29 +12,22 @@ import os
 import re
 import signal
 import subprocess
-import select
 import difflib
+import unicodedata
 
 META_INFO = {
-    'version'     : '1.2',
+    'version'     : '1.3',
     'license'     : 'BSD-3',
     'author'      : 'Matt Wang',
-    'email'       : 'mattwyl(@)gmail(.)com',
     'url'         : 'https://github.com/ymattw/ydiff',
     'keywords'    : 'colored incremental side-by-side diff',
     'description' : ('View colored, incremental diff in a workspace or from '
                      'stdin, with side by side and auto pager support')
 }
 
-if sys.hexversion < 0x02050000:
-    raise SystemExit('*** Requires python >= 2.5.0')    # pragma: no cover
+if sys.hexversion < 0x02060000:
+    raise SystemExit('*** Requires python >= 2.6.0')    # pragma: no cover
 
-# Python < 2.6 does not have next()
-try:
-    next
-except NameError:
-    def next(obj):
-        return obj.next()
 
 try:
     unicode
@@ -119,19 +112,19 @@ def strsplit(text, width):
     appended with the resetting sequence, and the second string is prefixed
     with all active colors.
     """
-    first = ""
-    second = ""
-    found_colors = ""
+    first = ''
+    second = ''
+    found_colors = ''
     chars_cnt = 0
     bytes_cnt = 0
     while text:
         append_len = 0
-        if text[0] == "\x1b":
-            color_end = text.find("m")
+        if text[0] == '\x1b':
+            color_end = text.find('m')
             if color_end != -1:
                 color = text[:color_end + 1]
-                if color == COLORS["reset"]:
-                    found_colors = ""
+                if color == COLORS['reset']:
+                    found_colors = ''
                 else:
                     found_colors += color
                 append_len = len(color)
@@ -142,7 +135,10 @@ def strsplit(text, width):
             # break if that string is already large enough.
             if chars_cnt >= width:
                 break
-            chars_cnt += 1
+            if unicodedata.east_asian_width(unicode(text[0])) in ("W", "F"):
+                chars_cnt += 2
+            else:
+                chars_cnt += 1
             append_len = 1
 
         first += text[:append_len]
@@ -289,119 +285,10 @@ class UnifiedDiff(object):
         return re.match('^Binary files .* differ$', line.rstrip())
 
 
-class PatchStream(object):
-
-    def __init__(self, diff_hdl):
-        self._diff_hdl = diff_hdl
-        self._stream_header_size = 0
-        self._stream_header = []
-
-        # Test whether stream is empty by read 1 line
-        line = self._diff_hdl.readline()
-        if not line:
-            self._is_empty = True
-        else:
-            self._stream_header.append(line)
-            self._stream_header_size += 1
-            self._is_empty = False
-
-    def is_empty(self):
-        return self._is_empty
-
-    def read_stream_header(self, stream_header_size):
-        """Returns a small chunk for patch type detect, suppose to call once"""
-        for i in range(1, stream_header_size):
-            line = self._diff_hdl.readline()
-            if not line:
-                break
-            self._stream_header.append(line)
-            self._stream_header_size += 1
-        return self._stream_header
-
-    def __iter__(self):
-        for line in self._stream_header:
-            yield line
-        try:
-            for line in self._diff_hdl:
-                yield line
-        except RuntimeError:
-            return
-
-
-class PatchStreamForwarder(object):
-    """A blocking stream forwarder use `select` and line buffered mode.  Feed
-    input stream to a diff format translator and read output stream from it.
-    Note input stream is non-seekable, and upstream has eaten some lines.
-    """
-    def __init__(self, istream, translator):
-        assert isinstance(istream, PatchStream)
-        assert isinstance(translator, subprocess.Popen)
-        self._istream = iter(istream)
-        self._in = translator.stdin
-        self._out = translator.stdout
-
-    def _can_read(self, timeout=0):
-        return select.select([self._out.fileno()], [], [], timeout)[0]
-
-    def _forward_line(self):
-        try:
-            line = next(self._istream)
-            self._in.write(line)
-        except StopIteration:
-            self._in.close()
-
-    def __iter__(self):
-        while True:
-            if self._can_read():
-                line = self._out.readline()
-                if line:
-                    yield line
-                else:
-                    return
-            elif not self._in.closed:
-                self._forward_line()
-
-
 class DiffParser(object):
 
     def __init__(self, stream):
-
-        header = [decode(line) for line in stream.read_stream_header(100)]
-        size = len(header)
-
-        if size >= 4 and (header[0].startswith('*** ') and
-                          header[1].startswith('--- ') and
-                          header[2].rstrip() == '***************' and
-                          header[3].startswith('*** ') and
-                          header[3].rstrip().endswith(' ****')):
-            # For context diff, try use `filterdiff` to translate it to unified
-            # format and provide a new stream
-            #
-            self._type = 'context'
-            try:
-                # Use line buffered mode so that to readline() in block mode
-                self._translator = subprocess.Popen(
-                    ['filterdiff', '--format=unified'], stdin=subprocess.PIPE,
-                    stdout=subprocess.PIPE, bufsize=1)
-            except OSError:
-                raise SystemExit('*** Context diff support depends on '
-                                 'filterdiff')
-            self._stream = PatchStreamForwarder(stream, self._translator)
-            return
-
-        for n in range(size):
-            if (header[n].startswith('--- ') and (n < size - 1) and
-                    header[n + 1].startswith('+++ ')):
-                self._type = 'unified'
-                self._stream = stream
-                break
-        else:
-            # `filterdiff` translates unknown diff to nothing, fall through to
-            # unified diff give ydiff a chance to show everything as headers
-            #
-            sys.stderr.write("*** unknown format, fall through to 'unified'\n")
-            self._type = 'unified'
-            self._stream = stream
+        self._stream = stream
 
     def get_diff_generator(self):
         """parse all diff lines, construct a list of UnifiedDiff objects"""
@@ -416,7 +303,6 @@ class DiffParser(object):
                 # is completed.  We yield previous diff if exists and construct
                 # a new one for this case.  Otherwise it's acutally an 'old'
                 # line starts with '--- '.
-                #
                 if (not diff._hunks or diff._hunks[-1].is_completed()):
                     if diff._old_path and diff._new_path and diff._hunks:
                         yield diff
@@ -453,7 +339,6 @@ class DiffParser(object):
             elif diff.is_only_in_dir(line) or diff.is_binary_differ(line):
                 # 'Only in foo:' and 'Binary files ... differ' are considered
                 # as separate diffs, so yield current diff, then this line
-                #
                 if diff._old_path and diff._new_path and diff._hunks:
                     # Current diff is comppletely constructed
                     yield diff
@@ -465,7 +350,6 @@ class DiffParser(object):
             else:
                 # All other non-recognized lines are considered as headers or
                 # hunk headers respectively
-                #
                 headers.append(line)
 
         # Validate and yield the last patch set if it is not yielded yet
@@ -479,7 +363,6 @@ class DiffParser(object):
         if headers:
             # Tolerate dangling headers, just yield a UnifiedDiff object with
             # only header lines
-            #
             yield UnifiedDiff(headers, '', '', [])
 
 
@@ -548,9 +431,7 @@ class DiffMarker(object):
                 # next stop modulo tab width
                 width = self._tab_width - (index - offset) % self._tab_width
                 line = line[:index] + ' ' * width + line[(index + 1):]
-            return (line
-                    .replace('\n', '')
-                    .replace('\r', ''))
+            return line.replace('\n', '').replace('\r', '')
 
         def _fit_with_marker_mix(text, base_color):
             """Wrap input text which contains mdiff tags, markup at the
@@ -578,12 +459,10 @@ class DiffMarker(object):
                     # u'\u554a' takes double width of a single letter, also
                     # this depends on your terminal font.  I guess audience of
                     # this tool never put that kind of symbol in their code :-)
-                    #
                     out.append(text[0])
                     text = text[1:]
 
             out.append(COLORS['reset'])
-
             return ''.join(out)
 
         # Set up number width, note last hunk might be empty
@@ -604,7 +483,6 @@ class DiffMarker(object):
                 # Each line is like 'nnn TEXT nnn TEXT\n', so width is half of
                 # [terminal size minus the line number columns and 3 separating
                 # spaces
-                #
                 width = (terminal_size()[0] - num_width * 2 - 3) // 2
             except Exception:
                 # If terminal detection failed, set back to default
@@ -739,17 +617,7 @@ class DiffMarker(object):
 
 
 def markup_to_pager(stream, opts):
-    """Pipe unified diff stream to pager (less).
-
-    Note: have to create pager Popen object before the translator Popen object
-    in PatchStreamForwarder, otherwise the `stdin=subprocess.PIPE` would cause
-    trouble to the translator pipe (select() never see EOF after input stream
-    ended), most likely python bug 12607 (http://bugs.python.org/issue12607)
-    which was fixed in python 2.7.3.
-
-    See issue #30 (https://github.com/ymattw/ydiff/issues/30) for more
-    information.
-    """
+    """Pipe unified diff stream to pager (less)."""
     pager_cmd = [opts.pager]
     pager_opts = (opts.pager_options.split(' ')
                   if opts.pager_options is not None
@@ -819,11 +687,25 @@ def terminal_size():
     return width, height
 
 
-def main():
-    if sys.platform != 'win32':
-        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
-    signal.signal(signal.SIGINT, signal.SIG_DFL)
+def trap_interrupts(entry_fn):
+    def entry_wrapper():
+        signal.signal(signal.SIGINT, signal.SIG_DFL)
+        if sys.platform != 'win32':
+            signal.signal(signal.SIGPIPE, signal.SIG_DFL)
+            return entry_fn()
+        else:
+            import errno
+            try:
+                return entry_fn()
+            except IOError as e:
+                if e.errno not in [errno.EPIPE, errno.EINVAL]:
+                    raise
+                return 0
+    return entry_wrapper
 
+
+@trap_interrupts
+def main():
     from optparse import (OptionParser, BadOptionError, AmbiguousOptionError,
                           OptionGroup)
 
@@ -887,21 +769,11 @@ def main():
 
     # Place possible options defined in YDIFF_OPTIONS at the beginning of argv
     ydiff_opts = [x for x in os.getenv('YDIFF_OPTIONS', '').split(' ') if x]
-
-    # TODO: Deprecate CDIFF_OPTIONS. Fall back to it and warn (for now).
-    if not ydiff_opts:
-        cdiff_opts = [x for x in os.getenv('CDIFF_OPTIONS', '').split(' ')
-                      if x]
-        if cdiff_opts:
-            sys.stderr.write('*** CDIFF_OPTIONS will be depreated soon, '
-                             'please use YDIFF_OPTIONS instead\n')
-            ydiff_opts = cdiff_opts
-
     opts, args = parser.parse_args(ydiff_opts + sys.argv[1:])
 
+    stream = None
     if not sys.stdin.isatty():
-        diff_hdl = (sys.stdin.buffer if hasattr(sys.stdin, 'buffer')
-                    else sys.stdin)
+        stream = getattr(sys.stdin, 'buffer', sys.stdin)
     else:
         vcs_name = revision_control_probe()
         if vcs_name is None:
@@ -911,34 +783,26 @@ def main():
             return 1
 
         if opts.log:
-            diff_hdl = revision_control_log(vcs_name, args)
-            if diff_hdl is None:
+            stream = revision_control_log(vcs_name, args)
+            if stream is None:
                 sys.stderr.write('*** %s does not support log command.\n' %
                                  vcs_name)
                 return 1
         else:
             # 'diff' is a must have feature.
-            diff_hdl = revision_control_diff(vcs_name, args)
-
-    stream = PatchStream(diff_hdl)
-
-    # Don't let empty diff pass thru
-    if stream.is_empty():
-        return 0
+            stream = revision_control_diff(vcs_name, args)
 
     if (opts.color == 'always' or
             (opts.color == 'auto' and sys.stdout.isatty())):
         markup_to_pager(stream, opts)
     else:
         # pipe out stream untouched to make sure it is still a patch
-        byte_output = (sys.stdout.buffer if hasattr(sys.stdout, 'buffer')
-                       else sys.stdout)
+        byte_output = getattr(sys.stdout, 'buffer', sys.stdout)
         for line in stream:
             byte_output.write(line)
 
-    if diff_hdl is not sys.stdin:
-        diff_hdl.close()
-
+    if stream is not None:
+        stream.close()
     return 0
 
 
